@@ -73,6 +73,7 @@ class SOMEIPHeader:
     """
     Represents a top-level SOMEIP packet (header and payload).
     """
+
     __format: typing.ClassVar[struct.Struct] = struct.Struct("!HHIHHBBBB")
     service_id: int
     method_id: int
@@ -80,9 +81,9 @@ class SOMEIPHeader:
     session_id: int
     interface_version: int
     message_type: SOMEIPMessageType
-    protocol_version: int = dataclasses.field(default=1)
-    return_code: SOMEIPReturnCode = dataclasses.field(default=SOMEIPReturnCode.E_OK)
-    payload: bytes = dataclasses.field(default=b"")
+    protocol_version: int = 1
+    return_code: SOMEIPReturnCode = SOMEIPReturnCode.E_OK
+    payload: bytes = b""
 
     @property
     def description(self):  # pragma: nocover
@@ -123,6 +124,9 @@ payload: {len(self.payload)} bytes"""
         except ValueError as exc:
             raise ParseError("bad someip return code {rc_b:#x}") from exc
 
+        if size < 8:
+            raise ParseError("SOMEIP length must be at least 8")
+
         return (
             size,
             lambda payload_b: cls(
@@ -210,6 +214,7 @@ class SOMEIPReader:
     Wrapper class around :class:`asyncio.StreamReader` that returns parsed
     :class:`SOMEIPHeader` from :meth:`read`
     """
+
     def __init__(self, reader: asyncio.StreamReader):
         self.reader = reader
 
@@ -269,6 +274,7 @@ class SOMEIPSDEntry:
     :param num_options_1: number of option (for unresolved options, run 1)
     :param num_options_2: number of option (for unresolved options, run 2)
     """
+
     __format: typing.ClassVar[struct.Struct] = struct.Struct("!BBBBHHBBHI")
     sd_type: SOMEIPSDEntryType
     service_id: int
@@ -277,13 +283,13 @@ class SOMEIPSDEntry:
     ttl: int
     minver_or_counter: int
 
-    options_1: typing.Tuple[SOMEIPSDOption, ...] = dataclasses.field(default=())
-    options_2: typing.Tuple[SOMEIPSDOption, ...] = dataclasses.field(default=())
+    options_1: typing.Tuple[SOMEIPSDOption, ...] = ()
+    options_2: typing.Tuple[SOMEIPSDOption, ...] = ()
 
-    option_index_1: typing.Optional[int] = dataclasses.field(default=None)
-    option_index_2: typing.Optional[int] = dataclasses.field(default=None)
-    num_options_1: typing.Optional[int] = dataclasses.field(default=None)
-    num_options_2: typing.Optional[int] = dataclasses.field(default=None)
+    option_index_1: typing.Optional[int] = None
+    option_index_2: typing.Optional[int] = None
+    num_options_1: typing.Optional[int] = None
+    num_options_2: typing.Optional[int] = None
 
     def __str__(self) -> str:  # pragma: nocover
         if self.sd_type in (
@@ -541,6 +547,7 @@ class SOMEIPSDOption:
     """
     Abstract base class representing SD options
     """
+
     __format: typing.ClassVar[struct.Struct] = struct.Struct("!HB")
     _options: typing.ClassVar[
         typing.Dict[int, typing.Type[SOMEIPSDAbstractOption]]
@@ -615,6 +622,7 @@ class SOMEIPSDUnknownOption(SOMEIPSDOption):
     :param type: the type identifier for this unknown option
     :param payload: the option payload
     """
+
     type: int
     payload: bytes
 
@@ -744,6 +752,7 @@ class L4Protocols(enum.IntEnum):
     """
     Enum for valid layer 4 protocol identifiers.
     """
+
     TCP = socket.IPPROTO_TCP
     UDP = socket.IPPROTO_UDP
 
@@ -754,8 +763,9 @@ class AbstractIPOption(SOMEIPSDAbstractOption, typing.Generic[T]):
     Abstract base class for options with IP payloads. Generalizes parsing and building
     based on :attr:`_format`, :attr:`_address_type` and :attr:`_family`.
     """
+
     _format: typing.ClassVar[struct.Struct]
-    _address_type: typing.ClassVar[typing.Type[T]]
+    _address_type: typing.ClassVar[typing.Type[typing.Any]]
     _family: typing.ClassVar[socket.AddressFamily]
     address: T
     l4proto: typing.Union[L4Protocols, int]
@@ -812,6 +822,7 @@ class EndpointOption(AbstractIPOption[T]):
     """
     Abstract base class for endpoint options (IPv4 or IPv6).
     """
+
     pass
 
 
@@ -819,6 +830,7 @@ class MulticastOption(AbstractIPOption[T]):
     """
     Abstract base class for multicast options (IPv4 or IPv6).
     """
+
     pass
 
 
@@ -826,6 +838,7 @@ class SDEndpointOption(AbstractIPOption[T]):
     """
     Abstract base class for SD Endpoint options (IPv4 or IPv6).
     """
+
     pass
 
 
@@ -833,6 +846,7 @@ class AbstractIPv4Option(AbstractIPOption[ipaddress.IPv4Address]):
     """
     Abstract base class for IPv4 options.
     """
+
     _format: typing.ClassVar[struct.Struct] = struct.Struct("!B4sBBH")
     _address_type = ipaddress.IPv4Address
     _family = socket.AF_INET
@@ -848,6 +862,7 @@ class AbstractIPv6Option(AbstractIPOption[ipaddress.IPv6Address]):
     """
     Abstract base class for IPv6 options.
     """
+
     _format: typing.ClassVar[struct.Struct] = struct.Struct("!B16sBBH")
     _address_type = ipaddress.IPv6Address
     _family = socket.AF_INET6
@@ -896,11 +911,12 @@ class SOMEIPSDHeader:
     """
     Represents a SOMEIP SD packet.
     """
+
     entries: typing.Tuple[SOMEIPSDEntry, ...]
-    options: typing.Tuple[SOMEIPSDOption, ...] = dataclasses.field(default=())
-    flag_reboot: bool = dataclasses.field(default=False)
-    flag_unicast: bool = dataclasses.field(default=True)
-    flags_unknown: int = dataclasses.field(default=0)
+    options: typing.Tuple[SOMEIPSDOption, ...] = ()
+    flag_reboot: bool = False
+    flag_unicast: bool = True
+    flags_unknown: int = 0
 
     def resolve_options(self):
         """
