@@ -7,6 +7,7 @@ import os
 import socket
 import struct
 import sys
+import typing
 import unittest
 import unittest.mock
 from dataclasses import replace
@@ -16,13 +17,19 @@ import someip.config as cfg
 import someip.sd as sd
 
 logging.captureWarnings(True)
-logging.basicConfig(level=logging.DEBUG)
+
+handler = logging.StreamHandler()
+handler.setFormatter(
+    logging.Formatter("%(asctime)s %(levelname)-10s %(name)s: %(message)s")
+)
+handler.setLevel(logging.DEBUG)
 
 logging.getLogger("asyncio").setLevel(logging.WARNING)
 logging.getLogger("someip").setLevel(logging.WARNING)
+logging.getLogger().addHandler(handler)
 
 
-PRECISION = 0.4 if os.environ.get('CI') == 'true' else 0.2  # in seconds
+PRECISION = 0.4 if os.environ.get("CI") == "true" else 0.2  # in seconds
 
 
 def ticks(n):
@@ -37,7 +44,9 @@ def setUpModule():
 # {{{ Utilities: pack_sd and _SendTiming
 def pack_sd(entries, reboot=True, unicast=True, session_id=1):
     msg = hdr.SOMEIPSDHeader(
-        flag_reboot=reboot, flag_unicast=unicast, entries=tuple(entries),
+        flag_reboot=reboot,
+        flag_unicast=unicast,
+        entries=tuple(entries),
     ).assign_option_indexes()
 
     data = hdr.SOMEIPHeader(
@@ -55,7 +64,7 @@ def pack_sd(entries, reboot=True, unicast=True, session_id=1):
 
 class _SendTiming(unittest.TestCase):
     def setup_timing(self, *mocks):
-        self.send_times = {}
+        self.send_times: typing.Dict[unittest.mock.Mock, typing.List[float]] = {}
 
         for mock in mocks:
             self.send_times[mock] = []
@@ -84,13 +93,16 @@ class _SendTiming(unittest.TestCase):
 
         # self.assertEqual(len(self.send_times), len(expected))
         for i, (expected_td, mock, expected_call) in enumerate(expected):
+            self.assertNotEqual(
+                len(mock.call_args_list), 0, msg=f"index {i} not called"
+            )
             actual_call = mock.call_args_list.pop(0)
             actual_td = tdiffs[mock].pop(0)
             self.assertEqual(actual_call, expected_call, msg=f"index {i} failed")
             self.assertLess(
                 abs(ticks(expected_td) - actual_td),
                 PRECISION,
-                msg=f"{ticks(expected_td)=} {actual_td=}",
+                msg=f"index {i} timing failed: {ticks(expected_td)=} {actual_td=}",
             )
 
         self.assertDictEqual(
@@ -99,7 +111,7 @@ class _SendTiming(unittest.TestCase):
 
 
 async def settle():
-    fut = asyncio.Future()
+    fut: asyncio.Future[None] = asyncio.Future()
     asyncio.get_event_loop().call_soon(fut.set_result, None)
     await fut
 
@@ -178,7 +190,9 @@ class TestSD(unittest.IsolatedAsyncioTestCase):
 
         # session_id wraps to 1 instead of 0
         r = itertools.chain(
-            range(1, 0x10000), range(0x100001, 0x20000), range(0x200001, 0x20020),
+            range(1, 0x10000),
+            range(0x10001, 0x20000),
+            range(0x20001, 0x20020),
         )
 
         for i in r:
@@ -356,7 +370,9 @@ class TestSD(unittest.IsolatedAsyncioTestCase):
 
         with self.assertLogs("someip", "ERROR"):
             prot.datagram_received(
-                replace(msg, method_id=0x1234).build(), self.fake_addr, multicast=False,
+                replace(msg, method_id=0x1234).build(),
+                self.fake_addr,
+                multicast=False,
             )
 
         prot.sd_message_received.assert_not_called()
@@ -458,7 +474,7 @@ class TestSD(unittest.IsolatedAsyncioTestCase):
             announce().method_calls,
             [
                 unittest.mock.call.handle_findservice(
-                    find_entry, self.fake_addr, False, True
+                    find_entry, self.fake_addr, False
                 ),
                 unittest.mock.call.handle_subscribe(subscribe_entry, self.fake_addr),
             ],
@@ -481,9 +497,7 @@ class TestSD(unittest.IsolatedAsyncioTestCase):
         self.assertEqual(
             announce().method_calls,
             [
-                unittest.mock.call.handle_findservice(
-                    find_entry, self.fake_addr, True, True
-                ),
+                unittest.mock.call.handle_findservice(find_entry, self.fake_addr, True),
             ],
         )
 
@@ -655,7 +669,19 @@ class TestSDDiscoveryTTLForever(_BaseSDDiscoveryTest):
         )
 
     async def test_sd_ignore(self):
-        self.prot.watcher_all_services.remove(self.mock)
+        self.prot.stop_watch_all_services(self.mock)
+
+        await settle()
+        self.assertCountEqual(
+            self.mock.method_calls,
+            [
+                unittest.mock.call.service_stopped(self.cfg_offer_5566, self.fake_addr),
+                unittest.mock.call.service_stopped(self.cfg_offer_5567, self.fake_addr),
+            ],
+        )
+
+        self.mock.reset_mock()
+
         offer_7777 = replace(self.offer_5566, service_id=0x7777)
         self.prot.handle_offer(offer_7777, self.fake_addr)
 
@@ -663,6 +689,94 @@ class TestSDDiscoveryTTLForever(_BaseSDDiscoveryTest):
 
         self.assertEqual(self.mock.method_calls, [])
 
+    async def test_watch_all_while_running(self):
+        newmock = unittest.mock.Mock()
+        self.prot.watch_all_services(newmock)
+        await settle()
+
+        self.assertEqual(
+            newmock.service_offered.call_args_list,
+            [
+                unittest.mock.call(self.cfg_offer_5566, self.fake_addr),
+                unittest.mock.call(self.cfg_offer_5567, self.fake_addr),
+            ],
+        )
+
+        newmock.service_stopped.assert_not_called()
+
+    async def test_stop_watch_all(self):
+        self.prot.stop_watch_all_services(self.mock)
+        await settle()
+
+        self.assertEqual(
+            self.mock.service_stopped.call_args_list,
+            [
+                unittest.mock.call(self.cfg_offer_5566, self.fake_addr),
+                unittest.mock.call(self.cfg_offer_5567, self.fake_addr),
+            ],
+        )
+
+        self.mock.reset_mock()
+
+        offer_5568 = replace(self.offer_5566, service_id=0x5568)
+        self.prot.handle_offer(offer_5568, self.fake_addr)
+        await settle()
+
+        self.assertEqual(self.mock.method_calls, [])
+
+    async def test_stop_watch_single(self):
+        self.prot.stop_watch_service(cfg.Service(service_id=0x5566), self.mock_single)
+        await settle()
+
+        self.assertEqual(
+            self.mock_single.service_stopped.call_args_list,
+            [
+                unittest.mock.call(self.cfg_offer_5566, self.fake_addr),
+            ],
+        )
+
+        self.mock_single.reset_mock()
+
+        # clear already known services to re-trigger callbacks
+        self.prot.found_services.stop_all()
+        await settle()
+        self.assertEqual(self.mock_single.method_calls, [])
+
+        # after found_services.stop_all(), new offer would trigger callbacks,
+        # but the listener is removed
+        self.prot.handle_offer(self.offer_5566, self.fake_addr)
+        await settle()
+        self.assertEqual(self.mock_single.method_calls, [])
+
+    async def test_watch_while_running(self):
+        self.prot.stop_watch_all_services(self.mock)
+        await settle()
+
+        newmock = unittest.mock.Mock()
+        self.prot.watch_service(cfg.Service(service_id=0x5566), newmock)
+        await settle()
+
+        self.assertEqual(
+            newmock.service_offered.call_args_list,
+            [unittest.mock.call(self.cfg_offer_5566, self.fake_addr)],
+        )
+
+        newmock.service_stopped.assert_not_called()
+
+        # 0x5567 was previously monitored from watch_all_services
+        # TTL is forever, so it's still in the found_services list
+        # so it will immediately trigger a service_offered callback
+        newmock = unittest.mock.Mock()
+        self.prot.watch_service(cfg.Service(service_id=0x5567), newmock)
+        await settle()
+
+        self.assertEqual(
+            newmock.service_offered.call_args_list,
+            [unittest.mock.call(self.cfg_offer_5567, self.fake_addr)],
+        )
+
+        newmock.service_stopped.assert_not_called()
+
 
 class TestSDDiscoveryTTL1(_BaseSDDiscoveryTest):
     TTL = 1
@@ -801,6 +915,107 @@ class TestSDDiscoveryTTL1(_BaseSDDiscoveryTest):
 
         self.assertEqual(self.mock_single.method_calls, [])
 
+    async def test_watch_all_while_running(self):
+        newmock = unittest.mock.Mock()
+        self.prot.watch_all_services(newmock)
+        await settle()
+
+        self.assertEqual(
+            newmock.service_offered.call_args_list,
+            [
+                unittest.mock.call(self.cfg_offer_5566, self.fake_addr),
+                unittest.mock.call(self.cfg_offer_5567, self.fake_addr),
+            ],
+        )
+
+        newmock.service_stopped.assert_not_called()
+
+    async def test_stop_watch_all(self):
+        self.prot.stop_watch_all_services(self.mock)
+        await settle()
+
+        self.assertEqual(
+            self.mock.service_stopped.call_args_list,
+            [
+                unittest.mock.call(self.cfg_offer_5566, self.fake_addr),
+                unittest.mock.call(self.cfg_offer_5567, self.fake_addr),
+            ],
+        )
+
+        self.mock.reset_mock()
+
+        # time out all offers
+        await asyncio.sleep(1.2)
+
+        offer_5568 = replace(self.offer_5566, service_id=0x5568)
+        self.prot.handle_offer(offer_5568, self.fake_addr)
+        self.prot.handle_offer(self.offer_5566, self.fake_addr)
+        await settle()
+
+        self.assertEqual(self.mock.method_calls, [])
+
+    async def test_stop_watch_single(self):
+        self.prot.stop_watch_service(cfg.Service(service_id=0x5566), self.mock_single)
+        await settle()
+
+        self.assertEqual(
+            self.mock_single.service_stopped.call_args_list,
+            [
+                unittest.mock.call(self.cfg_offer_5566, self.fake_addr),
+            ],
+        )
+
+        self.mock_single.reset_mock()
+
+        # time out all offers
+        await asyncio.sleep(1.2)
+        self.assertEqual(self.mock_single.method_calls, [])
+
+        # after found_services.stop_all(), new offer would trigger callbacks,
+        # but the listener is removed
+        self.prot.handle_offer(self.offer_5566, self.fake_addr)
+        await settle()
+        self.assertEqual(self.mock_single.method_calls, [])
+
+    async def test_watch_while_running(self):
+        self.prot.stop_watch_all_services(self.mock)
+        await settle()
+
+        # 0x5566 is still monitored from watch_service(0x5566)
+        newmock = unittest.mock.Mock()
+        self.prot.watch_service(cfg.Service(service_id=0x5566), newmock)
+        await settle()
+
+        self.assertEqual(
+            newmock.service_offered.call_args_list,
+            [unittest.mock.call(self.cfg_offer_5566, self.fake_addr)],
+        )
+
+        newmock.service_stopped.assert_not_called()
+
+        # 0x5567 is not monitored since stop_watch_all_services()
+        # TTL is 1, therefore the service will expire from found_services
+        await asyncio.sleep(0.6)
+        self.prot.handle_offer(self.offer_5567, self.fake_addr)
+        await asyncio.sleep(0.6)
+        self.prot.handle_offer(self.offer_5567, self.fake_addr)
+        await settle()
+
+        newmock = unittest.mock.Mock()
+        self.prot.watch_service(cfg.Service(service_id=0x5567), newmock)
+        await settle()
+
+        newmock.service_offered.assert_not_called()
+        newmock.service_stopped.assert_not_called()
+
+        self.prot.handle_offer(self.offer_5567, self.fake_addr)
+        await settle()
+
+        self.assertEqual(
+            newmock.service_offered.call_args_list,
+            [unittest.mock.call(self.cfg_offer_5567, self.fake_addr)],
+        )
+
 
 # }}}
 
@@ -976,17 +1191,17 @@ class TestSDFind(unittest.IsolatedAsyncioTestCase, _SendTiming):
     async def test_find_subscribe(self):
         local_addr = ("2001:db8::1111", 3333, 0, 0)
 
-        self.prot.find_subscribe_eventgroup(
-            cfg.Eventgroup(
-                service_id=0x5566,
-                instance_id=0xFFFF,
-                major_version=0xFF,
-                eventgroup_id=0x0000AAAA,
-                sockname=local_addr,
-                protocol=hdr.L4Protocols.UDP,
-            )
+        cfg_eventgroup = cfg.Eventgroup(
+            service_id=0x5566,
+            instance_id=0xFFFF,
+            major_version=0xFF,
+            eventgroup_id=0x0000AAAA,
+            sockname=local_addr,
+            protocol=hdr.L4Protocols.UDP,
         )
 
+        self.prot.find_subscribe_eventgroup(cfg_eventgroup)
+
         t = asyncio.create_task(self.prot.send_find_services())
 
         await asyncio.sleep(ticks(2.5))
@@ -1009,6 +1224,7 @@ class TestSDFind(unittest.IsolatedAsyncioTestCase, _SendTiming):
 
         self.prot.handle_offer(data, self.fake_addr)
 
+        # let offer time out
         await asyncio.sleep(ticks(10.1))
         await t
 
@@ -1029,6 +1245,11 @@ class TestSDFind(unittest.IsolatedAsyncioTestCase, _SendTiming):
             protocol=hdr.L4Protocols.UDP,
         )
 
+        self.prot.handle_offer(data, self.fake_addr)
+        await asyncio.sleep(ticks(2))
+        self.prot.stop_find_subscribe_eventgroup(cfg_eventgroup)
+        await asyncio.sleep(ticks(1))
+
         self.assertTiming(
             (2, self._mock_send_sd, unittest.mock.call([find])),
             (
@@ -1039,7 +1260,17 @@ class TestSDFind(unittest.IsolatedAsyncioTestCase, _SendTiming):
             (
                 12.5,
                 self._mock_subscribe_stop,
-                unittest.mock.call(evgrp, self.fake_addr, send=False),
+                unittest.mock.call(evgrp, self.fake_addr),
+            ),
+            (
+                12.5,
+                self._mock_subscribe_start,
+                unittest.mock.call(evgrp, self.fake_addr),
+            ),
+            (
+                14.5,
+                self._mock_subscribe_stop,
+                unittest.mock.call(evgrp, self.fake_addr),
             ),
         )
 
@@ -1052,7 +1283,9 @@ class TestSubscribeEventgroup(unittest.IsolatedAsyncioTestCase):
     local_addr = ("2001:db8::ff", 30331, 0, 0)
     remote1_addr = ("2001:db8::1", 30332, 0, 0)
     remote2_addr = ("2001:db8::2", 30337, 0, 0)
-    maxDiff = None
+    maxDiff: typing.Optional[int] = None
+    TTL = typing.ClassVar[int]
+    REFRESH_INTERVAL = typing.ClassVar[float]
 
     async def asyncSetUp(self):  # noqa: N802
         mock_sd = unittest.mock.Mock(spec_set=["timings", "log", "send_sd"])
@@ -1121,13 +1354,23 @@ class TestSubscribeEventgroupTTL3(TestSubscribeEventgroup):
         # add new eventgroup, show subscription in next cycle
         self.prot.subscribe_eventgroup(self.evgrp_3, self.remote1_addr)
 
+        await asyncio.sleep(0.1)
+
+        self.assertCountEqual(
+            self._mock_send_sd.call_args_list,
+            (unittest.mock.call([self.sub_evgrp_3], remote=self.remote1_addr),),
+        )
+        self._mock_send_sd.reset_mock()
+
+        # wait one refresh cycle
         await asyncio.sleep(1)
 
         self.assertCountEqual(
             self._mock_send_sd.call_args_list,
             (
                 unittest.mock.call(
-                    [self.sub_evgrp_1, self.sub_evgrp_3], remote=self.remote1_addr,
+                    [self.sub_evgrp_1, self.sub_evgrp_3],
+                    remote=self.remote1_addr,
                 ),
                 unittest.mock.call([self.sub_evgrp_2], remote=self.remote2_addr),
             ),
@@ -1140,7 +1383,8 @@ class TestSubscribeEventgroupTTL3(TestSubscribeEventgroup):
         await settle()
 
         self._mock_send_sd.assert_called_once_with(
-            [self.stop_sub_evgrp_1], remote=self.remote1_addr,
+            [self.stop_sub_evgrp_1],
+            remote=self.remote1_addr,
         )
         self._mock_send_sd.reset_mock()
 
@@ -1176,7 +1420,8 @@ class TestSubscribeEventgroupTTL3(TestSubscribeEventgroup):
         await asyncio.sleep(1)
 
         self._mock_send_sd.assert_called_once_with(
-            [self.sub_evgrp_2], remote=self.remote2_addr,
+            [self.sub_evgrp_2],
+            remote=self.remote2_addr,
         )
         self._mock_send_sd.reset_mock()
 
@@ -1205,7 +1450,8 @@ class TestSubscribeEventgroupTTLForever(TestSubscribeEventgroup):
 
         await settle()
         self._mock_send_sd.assert_called_once_with(
-            [self.sub_evgrp_3], remote=self.remote1_addr,
+            [self.sub_evgrp_3],
+            remote=self.remote1_addr,
         )
 
     async def test_sd_stop_send_stop(self):
@@ -1214,7 +1460,8 @@ class TestSubscribeEventgroupTTLForever(TestSubscribeEventgroup):
         await settle()
 
         self._mock_send_sd.assert_called_once_with(
-            [self.sub_evgrp_1], remote=self.remote1_addr,
+            [self.sub_evgrp_1],
+            remote=self.remote1_addr,
         )
         self._mock_send_sd.reset_mock()
 
@@ -1223,7 +1470,8 @@ class TestSubscribeEventgroupTTLForever(TestSubscribeEventgroup):
         await settle()
 
         self._mock_send_sd.assert_called_once_with(
-            [self.stop_sub_evgrp_1], remote=self.remote1_addr,
+            [self.stop_sub_evgrp_1],
+            remote=self.remote1_addr,
         )
 
     async def test_sd_stop_no_send_stop(self):
@@ -1232,7 +1480,8 @@ class TestSubscribeEventgroupTTLForever(TestSubscribeEventgroup):
         await settle()
 
         self._mock_send_sd.assert_called_once_with(
-            [self.sub_evgrp_1], remote=self.remote1_addr,
+            [self.sub_evgrp_1],
+            remote=self.remote1_addr,
         )
         self._mock_send_sd.reset_mock()
 
@@ -1265,6 +1514,7 @@ class TestSDAnnounce(unittest.IsolatedAsyncioTestCase, _SendTiming):
             REPETITIONS_BASE_DELAY=ticks(1),
             REQUEST_RESPONSE_DELAY_MIN=0,
             REQUEST_RESPONSE_DELAY_MAX=0,
+            SEND_COLLECTION_TIMEOUT=ticks(0.02),
         )
         mock_sd.log = logging.getLogger("someip.sd")
         mock_sd.send_sd = unittest.mock.Mock()
@@ -1300,12 +1550,20 @@ class TestSDAnnounce(unittest.IsolatedAsyncioTestCase, _SendTiming):
             options_1=(self.ep_2,),
         )
 
-        self.cfg_service_9999 = cfg.Service(service_id=0x9999,)
+        self.cfg_service_9999 = cfg.Service(
+            service_id=0x9999,
+        )
 
         self.listener_5566 = unittest.mock.Mock()
         self.listener_2233 = unittest.mock.Mock()
-        self.prot.announce_service(self.cfg_service_5566, self.listener_5566)
-        self.prot.announce_service(self.cfg_service_2233, self.listener_2233)
+        self.inst_5566 = sd.ServiceInstance(
+            self.cfg_service_5566, self.listener_5566, self.prot, mock_sd.timings
+        )
+        self.inst_2233 = sd.ServiceInstance(
+            self.cfg_service_2233, self.listener_2233, self.prot, mock_sd.timings
+        )
+        self.prot.announce_service(self.inst_5566)
+        self.prot.announce_service(self.inst_2233)
 
     async def test_announce_non_cyclic(self):
         self.prot.timings.REPETITIONS_MAX = 2
@@ -1327,14 +1585,17 @@ class TestSDAnnounce(unittest.IsolatedAsyncioTestCase, _SendTiming):
         )
 
         self.assertTiming(
+            # initial after INITIAL_DELAY
             (1, self._mock_send_sd, call),
+            # repetition 1 after 2**0 * REPETITIONS_BASE_DELAY == 1
             (2, self._mock_send_sd, call),
+            # repetition 2 after 2**1 * REPETITIONS_BASE_DELAY == 2
             (4, self._mock_send_sd, call),
         )
 
         self.reset_mock()
 
-        await asyncio.sleep(ticks(0.01))
+        await asyncio.sleep(ticks(0.03))
         self._mock_send_sd.assert_called_once_with(
             [
                 self.cfg_service_5566.create_offer_entry(0),
@@ -1360,17 +1621,23 @@ class TestSDAnnounce(unittest.IsolatedAsyncioTestCase, _SendTiming):
         )
 
         self.assertTiming(
+            # initial after INITIAL_DELAY
             (1, self._mock_send_sd, call),
+            # repetition 1 after 2**0 * REPETITIONS_BASE_DELAY == 1
             (2, self._mock_send_sd, call),
+            # repetition 2 after 2**1 * REPETITIONS_BASE_DELAY == 2
             (4, self._mock_send_sd, call),
+            # repetition 3 after 2**2 * REPETITIONS_BASE_DELAY == 4
             (8, self._mock_send_sd, call),
+            # cyclic offer after CYCLIC_OFFER_DELAY
             (10, self._mock_send_sd, call),
+            # cyclic offer after CYCLIC_OFFER_DELAY
             (12, self._mock_send_sd, call),
         )
 
         self.reset_mock()
 
-        await asyncio.sleep(ticks(0.01))
+        await asyncio.sleep(ticks(0.03))
         self._mock_send_sd.assert_called_once_with(
             [
                 self.cfg_service_5566.create_offer_entry(0),
@@ -1395,17 +1662,13 @@ class TestSDAnnounce(unittest.IsolatedAsyncioTestCase, _SendTiming):
     async def test_find_service_initial(self):
         data = self.cfg_service_5566.create_find_entry(self.TTL)
 
-        await self.prot.handle_findservice(
-            data, self.fake_addr, received_over_multicast=True, unicast_supported=True
-        )
+        self.prot.handle_findservice(data, self.fake_addr, received_over_multicast=True)
 
         self.assertTiming()
 
         self.prot.start()
         # sleep until middle of initial wait phase
-        await self.prot.handle_findservice(
-            data, self.fake_addr, received_over_multicast=True, unicast_supported=True
-        )
+        self.prot.handle_findservice(data, self.fake_addr, received_over_multicast=True)
         await asyncio.sleep(ticks(0.5))
         self.prot.stop()
 
@@ -1418,9 +1681,7 @@ class TestSDAnnounce(unittest.IsolatedAsyncioTestCase, _SendTiming):
         self.reset_mock()
 
         data = self.cfg_service_9999.create_find_entry()
-        await self.prot.handle_findservice(
-            data, self.fake_addr, received_over_multicast=True, unicast_supported=True
-        )
+        self.prot.handle_findservice(data, self.fake_addr, received_over_multicast=True)
 
         self.prot.stop()
 
@@ -1435,19 +1696,18 @@ class TestSDAnnounce(unittest.IsolatedAsyncioTestCase, _SendTiming):
         await asyncio.sleep(ticks(1))
         self.reset_mock()
 
-        await self.prot.handle_findservice(
+        self.prot.handle_findservice(
             self.cfg_service_5566.create_find_entry(),
             self.fake_addr,
             received_over_multicast=True,
-            unicast_supported=True,
         )
-        await self.prot.handle_findservice(
+        self.prot.handle_findservice(
             self.cfg_service_9999.create_find_entry(),
             self.fake_addr,
             received_over_multicast=True,
-            unicast_supported=True,
         )
 
+        await asyncio.sleep(ticks(1.03))
         self.prot.stop()
 
         self.assertTiming(
@@ -1470,19 +1730,18 @@ class TestSDAnnounce(unittest.IsolatedAsyncioTestCase, _SendTiming):
         await asyncio.sleep(ticks(1))
         self.reset_mock()
 
-        await self.prot.handle_findservice(
+        self.prot.handle_findservice(
             self.cfg_service_5566.create_find_entry(),
             self.fake_addr,
             received_over_multicast=False,
-            unicast_supported=True,
         )
-        await self.prot.handle_findservice(
+        self.prot.handle_findservice(
             self.cfg_service_9999.create_find_entry(),
             self.fake_addr,
             received_over_multicast=False,
-            unicast_supported=True,
         )
 
+        await asyncio.sleep(ticks(0.03))
         self.prot.stop()
 
         self.assertTiming(
@@ -1496,71 +1755,34 @@ class TestSDAnnounce(unittest.IsolatedAsyncioTestCase, _SendTiming):
             ),
         )
 
-    async def test_find_service_known_answer_multicast_supported(self):
+    async def test_announce_stop_single_service(self):
         self.prot.timings.INITIAL_DELAY_MIN = self.prot.timings.INITIAL_DELAY_MAX = 0
-        self.prot.timings.REQUEST_RESPONSE_DELAY_MIN = ticks(1)
-        self.prot.timings.REQUEST_RESPONSE_DELAY_MAX = ticks(1)
+        self.prot.timings.CYCLIC_OFFER_DELAY = ticks(1)
 
         self.prot.start()
-        await asyncio.sleep(ticks(1))
+        await asyncio.sleep(ticks(0.9))
         self.reset_mock()
 
-        await self.prot.handle_findservice(
-            self.cfg_service_5566.create_find_entry(),
-            self.fake_addr,
-            received_over_multicast=True,
-            unicast_supported=False,
-        )
-        await self.prot.handle_findservice(
-            self.cfg_service_9999.create_find_entry(),
-            self.fake_addr,
-            received_over_multicast=True,
-            unicast_supported=False,
-        )
-
-        self.prot.stop()
-
+        self.prot.stop_announce_service(self.inst_5566)
+        await asyncio.sleep(ticks(0.03))
         self.assertTiming(
             (
-                2,
+                0.9,
                 self._mock_send_sd,
                 unittest.mock.call(
-                    [self.cfg_service_5566.create_offer_entry(self.TTL)], remote=None
+                    [self.cfg_service_5566.create_offer_entry(0)], remote=None
                 ),
             ),
         )
-
-    async def test_find_service_known_answer_multicast_delay(self):
-        self.prot.timings.INITIAL_DELAY_MIN = self.prot.timings.INITIAL_DELAY_MAX = 0
-        self.prot.timings.REQUEST_RESPONSE_DELAY_MIN = ticks(1)
-        self.prot.timings.REQUEST_RESPONSE_DELAY_MAX = ticks(1)
-
-        self.prot.start()
-        await asyncio.sleep(ticks(11.5))
         self.reset_mock()
 
-        await self.prot.handle_findservice(
-            self.cfg_service_5566.create_find_entry(),
-            self.fake_addr,
-            received_over_multicast=True,
-            unicast_supported=True,
-        )
-        await self.prot.handle_findservice(
-            self.cfg_service_9999.create_find_entry(),
-            self.fake_addr,
-            received_over_multicast=True,
-            unicast_supported=True,
-        )
-
-        await asyncio.sleep(ticks(1.1))
-        self.prot.stop()
-
+        await asyncio.sleep(ticks(0.2))
         self.assertTiming(
             (
-                12.5,
+                1.0,
                 self._mock_send_sd,
                 unittest.mock.call(
-                    [self.cfg_service_5566.create_offer_entry(self.TTL)], remote=None
+                    [self.cfg_service_2233.create_offer_entry(self.TTL)], remote=None
                 ),
             ),
         )
@@ -1577,7 +1799,11 @@ class _BaseSDSubscriptionTest(unittest.IsolatedAsyncioTestCase):
 
     async def asyncSetUp(self):  # noqa: N802
         mock_sd = unittest.mock.Mock(spec_set=["timings", "log", "send_sd"])
-        mock_sd.timings = sd.Timings()
+        mock_sd.timings = sd.Timings(
+            INITIAL_DELAY_MIN=0,
+            INITIAL_DELAY_MAX=0,
+            REPETITIONS_MAX=0,
+        )
         mock_sd.log = logging.getLogger("someip.sd")
         mock_sd.send_sd = unittest.mock.Mock()
         self._mock_send_sd = mock_sd.send_sd
@@ -1620,12 +1846,23 @@ class _BaseSDSubscriptionTest(unittest.IsolatedAsyncioTestCase):
         self.stop_subscribe_5566_3333 = replace(self.subscribe_5566_3333, ttl=0)
 
         self.mock = unittest.mock.Mock()
-        self.prot.announce_service(self.cfg_5566, self.mock._5566)
-        self.prot.announce_service(self.cfg_5567, self.mock._5567)
+        self.inst_5566 = sd.ServiceInstance(
+            self.cfg_5566, self.mock._5566, self.prot, mock_sd.timings
+        )
+        self.inst_5567 = sd.ServiceInstance(
+            self.cfg_5567, self.mock._5567, self.prot, mock_sd.timings
+        )
+        self.prot.announce_service(self.inst_5566)
+        self.prot.announce_service(self.inst_5567)
 
-        await self.prot.handle_subscribe(self.subscribe_5566_3333, self.fake_sd_addr)
+        self.prot.start()
 
-        await settle()
+        await asyncio.sleep(ticks(0.03))
+        self.reset_mock()
+
+        self.prot.handle_subscribe(self.subscribe_5566_3333, self.fake_sd_addr)
+
+        await asyncio.sleep(ticks(0.03))
 
         self._mock_send_sd.assert_called_once_with(
             [self.ack_5566_3333], remote=self.fake_sd_addr
@@ -1650,15 +1887,13 @@ class _BaseSDSubscriptionTest(unittest.IsolatedAsyncioTestCase):
 
     async def _test_sd_reject_subscription(self):
         self.mock._5566.client_subscribed.side_effect = sd.NakSubscription
-        await self.prot.handle_subscribe(
-            self.stop_subscribe_5566_3333, self.fake_sd_addr
-        )
-        await settle()
+        self.prot.handle_subscribe(self.stop_subscribe_5566_3333, self.fake_sd_addr)
+        await asyncio.sleep(ticks(0.03))
 
         self.reset_mock()
 
-        await self.prot.handle_subscribe(self.subscribe_5566_3333, self.fake_sd_addr)
-        await settle()
+        self.prot.handle_subscribe(self.subscribe_5566_3333, self.fake_sd_addr)
+        await asyncio.sleep(ticks(0.03))
 
         self._mock_send_sd.assert_called_once_with(
             [self.nack_5566_3333], remote=self.fake_sd_addr
@@ -1676,8 +1911,8 @@ class _BaseSDSubscriptionTest(unittest.IsolatedAsyncioTestCase):
         self.reset_mock()
         self.mock._5566.client_subscribed.side_effect = None
 
-        await self.prot.handle_subscribe(self.subscribe_5566_3333, self.fake_sd_addr)
-        await settle()
+        self.prot.handle_subscribe(self.subscribe_5566_3333, self.fake_sd_addr)
+        await asyncio.sleep(ticks(0.03))
 
         self._mock_send_sd.assert_called_once_with(
             [self.ack_5566_3333], remote=self.fake_sd_addr
@@ -1692,6 +1927,27 @@ class _BaseSDSubscriptionTest(unittest.IsolatedAsyncioTestCase):
             ],
         )
 
+    async def _test_sd_stop_and_subscribe(self):
+        self.prot.handle_subscribe(self.stop_subscribe_5566_3333, self.fake_sd_addr)
+        self.prot.handle_subscribe(self.subscribe_5566_3333, self.fake_sd_addr)
+        await asyncio.sleep(ticks(0.03))
+
+        self._mock_send_sd.assert_called_once_with(
+            [self.ack_5566_3333], remote=self.fake_sd_addr
+        )
+
+        self.assertEqual(
+            self.mock.method_calls,
+            [
+                unittest.mock.call._5566.client_unsubscribed(
+                    self.subscription_5566_3333, self.fake_sd_addr
+                ),
+                unittest.mock.call._5566.client_subscribed(
+                    self.subscription_5566_3333, self.fake_sd_addr
+                ),
+            ],
+        )
+
     async def _test_sd_disconnect(self):
         self.prot.connection_lost(None)
 
@@ -1706,16 +1962,32 @@ class _BaseSDSubscriptionTest(unittest.IsolatedAsyncioTestCase):
             ],
         )
 
+    async def _test_sd_stopping_service(self):
+        self.prot.stop_announce_service(self.inst_5566)
+
+        await asyncio.sleep(ticks(0.03))
+
+        self.assertEqual(
+            self.mock.method_calls,
+            [
+                unittest.mock.call._5566.client_unsubscribed(
+                    self.subscription_5566_3333, self.fake_sd_addr
+                ),
+            ],
+        )
+
 
 class TestSDSubscriptionTTLForever(_BaseSDSubscriptionTest):
     TTL = sd.TTL_FOREVER
 
     test_sd_reject_subscription = _BaseSDSubscriptionTest._test_sd_reject_subscription
     test_sd_disconnect = _BaseSDSubscriptionTest._test_sd_disconnect
+    test_sd_stopping_service = _BaseSDSubscriptionTest._test_sd_stopping_service
+    test_sd_stop_and_subscribe = _BaseSDSubscriptionTest._test_sd_stop_and_subscribe
 
     async def test_sd_reboot_ttl_forever(self):
-        await self.prot.handle_subscribe(self.subscribe_5566_3333, self.fake_sd_addr)
-        await settle()
+        self.prot.handle_subscribe(self.subscribe_5566_3333, self.fake_sd_addr)
+        await asyncio.sleep(ticks(0.03))
 
         self._mock_send_sd.assert_called_once_with(
             [self.ack_5566_3333], remote=self.fake_sd_addr
@@ -1724,8 +1996,8 @@ class TestSDSubscriptionTTLForever(_BaseSDSubscriptionTest):
         self.mock.assert_not_called()
         self.reset_mock()
 
-        await self.prot.handle_subscribe(self.subscribe_5566_3333, self.fake_sd_addr)
-        await settle()
+        self.prot.handle_subscribe(self.subscribe_5566_3333, self.fake_sd_addr)
+        await asyncio.sleep(ticks(0.03))
 
         self._mock_send_sd.assert_called_once_with(
             [self.ack_5566_3333], remote=self.fake_sd_addr
@@ -1736,8 +2008,8 @@ class TestSDSubscriptionTTLForever(_BaseSDSubscriptionTest):
 
         self.prot.reboot_detected(self.fake_sd_addr)
         await settle()
-        await self.prot.handle_subscribe(self.subscribe_5566_3333, self.fake_sd_addr)
-        await settle()
+        self.prot.handle_subscribe(self.subscribe_5566_3333, self.fake_sd_addr)
+        await asyncio.sleep(ticks(0.03))
 
         self._mock_send_sd.assert_called_once_with(
             [self.ack_5566_3333], remote=self.fake_sd_addr
@@ -1759,10 +2031,8 @@ class TestSDSubscriptionTTLForever(_BaseSDSubscriptionTest):
 
         # send stopsubscribe, should reach listener
 
-        await self.prot.handle_subscribe(
-            self.stop_subscribe_5566_3333, self.fake_sd_addr
-        )
-        await settle()
+        self.prot.handle_subscribe(self.stop_subscribe_5566_3333, self.fake_sd_addr)
+        await asyncio.sleep(ticks(0.03))
 
         self._mock_send_sd.assert_not_called()
 
@@ -1776,20 +2046,19 @@ class TestSDSubscriptionTTLForever(_BaseSDSubscriptionTest):
         )
 
     async def test_sd_multiple_listener(self):
-        await self.prot.handle_subscribe(
-            self.stop_subscribe_5566_3333, self.fake_sd_addr
-        )
-        await settle()
+        self.prot.handle_subscribe(self.stop_subscribe_5566_3333, self.fake_sd_addr)
+        await asyncio.sleep(ticks(0.03))
 
         self.reset_mock()
 
-        self.prot.announce_service(self.cfg_5566, self.mock._5566_2)
+        inst_5566_2 = sd.ServiceInstance(
+            self.cfg_5566, self.mock._5566_2, self.prot, self.prot.timings
+        )
+        self.prot.announce_service(inst_5566_2)
 
         with self.assertLogs("someip.sd.announce", "WARNING") as cm:
-            await self.prot.handle_subscribe(
-                self.subscribe_5566_3333, self.fake_sd_addr
-            )
-            await settle()
+            self.prot.handle_subscribe(self.subscribe_5566_3333, self.fake_sd_addr)
+            await asyncio.sleep(ticks(0.03))
         self.assertTrue(any("multiple configured" in msg for msg in cm.output))
 
         self.assertEqual(
@@ -1798,6 +2067,9 @@ class TestSDSubscriptionTTLForever(_BaseSDSubscriptionTest):
                 unittest.mock.call._5566.client_subscribed(
                     self.subscription_5566_3333, self.fake_sd_addr
                 ),
+                unittest.mock.call._5566_2.client_subscribed(
+                    self.subscription_5566_3333, self.fake_sd_addr
+                ),
             ],
         )
 
@@ -1807,12 +2079,14 @@ class TestSDSubscriptionTTL1(_BaseSDSubscriptionTest):
 
     test_sd_reject_subscription = _BaseSDSubscriptionTest._test_sd_reject_subscription
     test_sd_disconnect = _BaseSDSubscriptionTest._test_sd_disconnect
+    test_sd_stopping_service = _BaseSDSubscriptionTest._test_sd_stopping_service
+    test_sd_stop_and_subscribe = _BaseSDSubscriptionTest._test_sd_stop_and_subscribe
 
     async def test_sd_reboot(self):
         self.prot.reboot_detected(self.fake_sd_addr)
         await settle()
-        await self.prot.handle_subscribe(self.subscribe_5566_3333, self.fake_sd_addr)
-        await settle()
+        self.prot.handle_subscribe(self.subscribe_5566_3333, self.fake_sd_addr)
+        await asyncio.sleep(ticks(0.03))
 
         self._mock_send_sd.assert_called_once_with(
             [self.ack_5566_3333], remote=self.fake_sd_addr
@@ -1839,8 +2113,8 @@ class TestSDSubscriptionTTL1(_BaseSDSubscriptionTest):
         self.mock.assert_not_called()
 
     async def test_sd(self):
-        await self.prot.handle_subscribe(self.subscribe_5566_3333, self.fake_sd_addr)
-        await settle()
+        self.prot.handle_subscribe(self.subscribe_5566_3333, self.fake_sd_addr)
+        await asyncio.sleep(ticks(0.03))
 
         self._mock_send_sd.assert_called_once_with(
             [self.ack_5566_3333], remote=self.fake_sd_addr
@@ -1850,10 +2124,8 @@ class TestSDSubscriptionTTL1(_BaseSDSubscriptionTest):
 
         await asyncio.sleep(0.2)
 
-        await self.prot.handle_subscribe(
-            self.stop_subscribe_5566_3333, self.fake_sd_addr
-        )
-        await settle()
+        self.prot.handle_subscribe(self.stop_subscribe_5566_3333, self.fake_sd_addr)
+        await asyncio.sleep(ticks(0.03))
 
         self.assertEqual(
             self.mock.method_calls,
@@ -1866,8 +2138,8 @@ class TestSDSubscriptionTTL1(_BaseSDSubscriptionTest):
 
         self.reset_mock()
 
-        await self.prot.handle_subscribe(self.subscribe_5566_3333, self.fake_sd_addr)
-        await settle()
+        self.prot.handle_subscribe(self.subscribe_5566_3333, self.fake_sd_addr)
+        await asyncio.sleep(ticks(0.03))
 
         self._mock_send_sd.assert_called_once_with(
             [self.ack_5566_3333], remote=self.fake_sd_addr
@@ -1900,19 +2172,15 @@ class TestSDSubscriptionTTL1(_BaseSDSubscriptionTest):
         # service already removed by timeout detection above.
         # send stopoffer again, should not reach listener
 
-        await self.prot.handle_subscribe(
-            self.stop_subscribe_5566_3333, self.fake_sd_addr
-        )
-        await settle()
+        self.prot.handle_subscribe(self.stop_subscribe_5566_3333, self.fake_sd_addr)
+        await asyncio.sleep(ticks(0.03))
 
         self.mock.assert_not_called()
 
     async def test_sd_unknown(self):
         with self.assertLogs("someip.sd.announce", "WARNING") as cm:
-            await self.prot.handle_subscribe(
-                self.subscribe_5566_1234, self.fake_sd_addr
-            )
-            await settle()
+            self.prot.handle_subscribe(self.subscribe_5566_1234, self.fake_sd_addr)
+            await asyncio.sleep(ticks(0.03))
 
         self.assertTrue(any("subscribe for unknown" in msg for msg in cm.output))
 
@@ -1928,7 +2196,7 @@ class TestSDSubscriptionTTL1(_BaseSDSubscriptionTest):
 
 # {{{ ServiceDiscoveryProtocol multicast endpoints
 class _MulticastEndpointsTest:
-    maxDiff = None
+    maxDiff: typing.Optional[int] = None
     AF: socket.AddressFamily
     bind_lo_addr: str
     bind_mc_addr: str
@@ -2142,5 +2410,10 @@ class TestMulticastEndpointsV6(
 # }}}
 
 
+if "unittest.util" in __import__("sys").modules:
+    # Show full diff in self.assertEqual.
+    __import__("sys").modules["unittest.util"]._MAX_LENGTH = 999999999
+
+
 if __name__ == "__main__":
     unittest.main()
